--[[
    Data miner for all minipets and mounts
    $Revision: 164 $
]]--

--[[
Copyright (c) 2008, LordFarlander
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
]]--

local http = require( "socket.http" );
local lastRequest = nil;

local ignoreTeleport =
{
      ["Blink"] = true,
      ["Arm of Law"] = true,
      ["Gate of Shazzrah"] = true,
      ["Scroll of Recall"] = true,
      ["Magic Pull"] = true,
      ["Shadow Pounce"] = true,
      ["Shadowstep"] = true,
      ["Summon"] = true,
      ["Summon Menagerie"] = true,
      ["Tactical Blink"] = true,
      ["Time Shift"] = true,
      ["Tunnel"] = true,
      ["Vanish"] = true,
      ["Warp"] = true,
      [39700] = true,
      [54643] = true,
};

persistence =
{
	store = function (path, ...)
		local f, e = io.open(path, "w");
		if f then
			f:write("-- Persistent Data\n");
			f:write("return ");
			persistence.write(f, select(1,...), 0);
			for i = 2, select("#", ...) do
				f:write(",\n");
				persistence.write(f, select(i,...), 0);
			end;
			f:write("\n");
		else
			error(e);
		end;
		f:close();
	end;
	
	load = function (path)
		local f, e = loadfile(path);
		if f then
			return f();
		else
			return nil, e;
			--error(e);
		end;
	end;
	
	write = function (f, item, level)
		local t = type(item);
		persistence.writers[t](f, item, level);
	end;
	
	writeIndent = function (f, level)
		for i = 1, level do
			f:write("\t");
		end;
	end;
	
	writers = {
		["nil"] = function (f, item, level)
				f:write("nil");
			end;
		["number"] = function (f, item, level)
				f:write(tostring(item));
			end;
		["string"] = function (f, item, level)
				f:write(string.format("%q", item));
			end;
		["boolean"] = function (f, item, level)
				if item then
					f:write("true");
				else
					f:write("false");
				end
			end;
		["table"] = function (f, item, level)
				f:write("{\n");
				for k, v in pairs(item) do
					persistence.writeIndent(f, level+1);
					f:write("[");
					persistence.write(f, k, level+1);
					f:write("] = ");
					persistence.write(f, v, level+1);
					f:write(";\n");
				end
				persistence.writeIndent(f, level);
				f:write("}");
			end;
		["function"] = function (f, item, level)
				-- Does only work for "normal" functions, not those
				-- with upvalues or c functions
				local dInfo = debug.getinfo(item, "uS");
				if dInfo.nups > 0 then
					f:write("nil -- functions with upvalue not supported\n");
				elseif dInfo.what ~= "Lua" then
					f:write("nil -- function is not a lua function\n");
				else
					local r, s = pcall(string.dump,item);
					if r then
						f:write(string.format("loadstring(%q)", s));
					else
						f:write("nil -- function could not be dumped\n");
					end
				end
			end;
		["thread"] = function (f, item, level)
				f:write("nil --thread\n");
			end;
		["userdata"] = function (f, item, level)
				f:write("nil --userdata\n");
			end;
	}
}

function GetPage( url )
    return http.request( url );
end--GetPage( url )

function ProcessMount( mount )
    if( not mount.Type ) then
        mount.Type = {};
    end--if
    if( not mount.GroundSpeed ) then
        mount.GroundSpeed = 100;
        mount.AquaticSpeed = 100;
        mount.FlyingSpeed = 0;
        mount.Type[1] = "Ground";
        mount.Type[2] = "Slow";
    end--if
    if( not mount.Passengers ) then
        mount.Passengers = 1;
    end--if
    if( ( mount.Passengers > 1 ) and ( mount.Type[1] ~= "Passenger" ) ) then
        mount.Type[3] = mount.Type[2];
        mount.Type[2] = mount.Type[1];
        mount.Type[1] = "Passenger";
    end--if
    mount.Value = ("%i|%i|%i|%i"):format( mount.GroundSpeed, mount.AquaticSpeed, mount.FlyingSpeed, mount.Passengers );
    if( mount.Skills ) then
        local first = true;

        mount.Value = mount.Value .. "|";
        for name, value in pairs( mount.Skills ) do
            if( first ) then
                first = false;
                mount.Value = mount.Value .. ("%s%i"):format( name:sub( 1, 1 ), value );
            else
                mount.Value = mount.Value .. ("/%s%i"):format( name:sub( 1, 1 ), value );
            end--if
        end--for
    end--if
    if( not mount.Family ) then
        if( mount.SpellName:find( "[Bb]ear" ) ) then
            mount.Family = { "Bear" };
		elseif( mount.SpellName:find( "Broom" ) ) then
		    mount.Family = { "Broom", "Enchanted" };
        elseif( mount.SpellName:find( "Carpet" ) ) then
            mount.Family = { "Carpet", "Enchanted" };
        elseif( mount.SpellName:find( "Charger" ) ) then
            mount.Family = { "Horse", "Class" };
        elseif( mount.SpellName:find( "Dragonhawk" ) ) then
            mount.Family = { "Dragonhawk" };            
        elseif( mount.SpellName:find( "Drake" ) ) then
            mount.Family = { "Dragon" };
            if( mount.SpellName:find( "Proto%-" ) ) then
                table.insert( mount.Family, "ProtoDrake" );
            elseif( mount.SpellName:find( "Nether" ) ) then
                table.insert( mount.Family, "NetherDrake" );
            else
                table.insert( mount.Family, "Drake" );
            end--if
        elseif( mount.SpellName:find( "Elekk" ) ) then
            mount.Family = { "Elekk" };
        elseif( mount.SpellName:find( "Frost ?[Ww]yrm" ) or mount.SpellName:find( "Frostbrood" ) ) then
            mount.Family = { "Dragon", "Skeletal", "FrostWyrm" };
        elseif( mount.SpellName:find( "Gryphon" ) ) then
            mount.Family = { "Gryphon" };
        elseif( mount.SpellName:find( "Hawkstrider" ) ) then
            mount.Family = { "Bird", "Hawkstrider" };            
        elseif( mount.SpellName:find( "Hippogryph" ) ) then
            mount.Family = { "Hippogryph" };
        elseif( mount.SpellName:find( "Skeletal" ) ) then
            mount.Family = { "Skeletal", "Horse" };
        elseif( mount.SpellName:find( "[Hh]orse" ) or mount.SpellName:find( "Steed" ) or mount.SpellName:find( "Stallion" ) or mount.SpellName:find( "Mare" ) ) then
            mount.Family = { "Horse" };
        elseif( mount.SpellName:find( "Kodo" ) ) then
            mount.Family = { "Kodo" };
        elseif( mount.SpellName:find( "Mammoth" ) ) then
            mount.Family = { "Mammoth" };
        elseif( mount.SpellName:find( "Mechanostrider" ) ) then
            mount.Family = { "Mechanical", "Mechanostrider", "Bird" };
        elseif( mount.SpellName:find( "Nether [Rr]ay" ) ) then
            mount.Family = { "NetherRay" };            
        elseif( mount.SpellName:find( "Qiraji" ) ) then
            mount.Family = { "Bug", "Qiraji" };
        elseif( mount.SpellName:find( "Ram" ) ) then
            mount.Family = { "Ram" };
        elseif( mount.SpellName:find( "Raptor" ) ) then
            mount.Family = { "Raptor" };
        elseif( mount.SpellName:find( "Rocket" ) ) then
            mount.Family = { "Mechanical", "Rocket" };			
        elseif( mount.SpellName:find( "[Ss]aber" ) or mount.SpellName:find( "Tiger" ) or mount.SpellName:find( "Panther" ) ) then
            mount.Family = { "Cat" };            
        elseif( mount.SpellName:find( "Talbuk" ) ) then
            mount.Family = { "Talbuk" };
        elseif( mount.SpellName:find( "Turtle" ) ) then
            mount.Family = { "Turtle" };
        elseif( mount.SpellName:find( "Wind ?[Rr]ider" ) ) then
            mount.Family = { "WindRider" };            
        elseif( mount.SpellName:find( "[Ww]olf" ) ) then
            mount.Family = { "Wolf" };
        elseif( mount.SpellName:find( "Zhevra" ) ) then
            mount.Family = { "Zhevra" };
        end--if
        if( not mount.Family ) then
            print( ("Could not find a class for '%s'"):format( mount.SpellName ) );
        end--if
    end--if
end--ProcessMount( mount )

function ProcessCritter( critter )
    if( not critter.Type ) then
        critter.Type = {};
    end--if
    if( not critter.Type[1] ) then
        critter.Type[1] = "Normal";
    end--if
    if( critter.InventorySlot ) then
        critter.Type[1] = "Equipment";
        if( critter.InventorySlot == "HEADSLOT" ) then
            critter.Type[2] = "Head";
        end--if
        critter.Value = critter.InventorySlot;
    end--if
end--ProcessCritter( critter )

function ProcessTranslocate( spell )
    if( spell ) then
        spell.Type = { [1] = "Teleport", [2] = "Translocate", };
        spell.ClassSpell = true;
    end--if
end--ProcessTranslocate( spell )

function ProcessMageTeleport( spell )
    if( spell ) then
        spell.Type = { [1] = "Teleport", [2] = "Mage", };
        spell.ClassSpell = true;
    end--if
end--ProcessMageTeleport( spell )

function ProcessDruidTeleport( spell )
    if( spell ) then
        spell.Type = { [1] = "Teleport", [2] = "Druid", };
        spell.ClassSpell = true;
    end--if
end--ProcessDruidTeleport( spell )

function ProcessHomeTeleport( spell )
    if( spell ) then
        spell.Type = { [1] = "Teleport", [2] = "Home", };
        spell.ClassSpell = true;
    end--if
end--ProcessHomeTeleport( spell )

function ProcessTeleport( spell )
    if( spell ) then
        spell.Type = { [1] = "Teleport", };
        if( ignoreTeleport[spell.SpellName] or ignoreTeleport[spell.Spell] ) then
            spell.Ignore = true;
        end--if
        spell.ClassSpell = true;
    end--if
end--ProcessTeleport( spell )

local HEADER = "--\[\[\n"..
    "    Database of all minipets and mounts\n"..
    "    $Revision: 164 $\n"..
    "\]\]--"..
    "\n"..
    "\n"..
    "--\[\[\nCopyright (c) 2008, LordFarlander\n"..
    "All rights reserved.\n"..
    "\n"..
    "Redistribution and use in source and binary forms, with or without\n"..
    "modification, are permitted provided that the following conditions are met:\n"..
    "\n"..
    "    * Redistributions of source code must retain the above copyright notice,\n"..
    "      this list of conditions and the following disclaimer.\n"..
    "    * Redistributions in binary form must reproduce the above copyright notice,\n"..
    "      this list of conditions and the following disclaimer in the documentation\n"..
    "      and/or other materials provided with the distribution.\n"..
    "\n"..
    "THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS \"AS IS\" AND\n"..
    "ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED\n"..
    "WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE\n"..
    "DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR\n"..
    "ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES\n"..
    "(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;\n"..
    "LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON\n"..
    "ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT\n"..
    "(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS\n"..
    "SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.\n"..
    "\]\]--\n"..
    "\n"..
    "if( not LibStub( \"LibPeriodicTable-3.1\", true ) ) then\n    error( \"PT3 must be loaded before data\" );\nend--if\n";

--local compatiblityVersion = 32000;
local latestVersion = 32000;

local data = { [latestVersion] = persistence.load( "datacache.lua" ) or { database = {}, } };
local latestData = data[latestVersion];
local compatibleData = nil;
local reverseLookup = {};
local urlLists = {
    Critters = {
        Spells = { "http://www.wowhead.com/?spells=-6", "http://ptr.wowhead.com/?spells=-6" },
        Items = { "http://www.wowhead.com/?items=15.2", "http://ptr.wowhead.com/?items=15.2" },
        Process = ProcessCritter,
    }, --Critters
    Mounts = {
        Spells = { "http://www.wowhead.com/?spells=-5", "http://ptr.wowhead.com/?spells=-5" },
        Items = { "http://www.wowhead.com/?items=15.5&filter=qu=0:1:2:3", "http://www.wowhead.com/?items=15.5&filter=qu=4:5:6:7", "http://www.wowhead.com/?items=15.-7",
                  "http://ptr.wowhead.com/?items=15.5&filter=qu=0:1:2:3", "http://ptr.wowhead.com/?items=15.5&filter=qu=4:5:6:7", "http://ptr.wowhead.com/?items=15.-7" },
        Process = ProcessMount,
    }, --Mounts
    Spells = {
        Teleport = {
            Translocate = {
                Spells = { "http://www.wowhead.com/?spells=0&filter=na=translocate", },
                Process = ProcessTranslocate,
            },
            Mage = {
                Spells = { "http://www.wowhead.com/?spells=7.8&filter=na=Teleport", },
                Process = ProcessMageTeleport,
            },
            Druid = {
                Spells = { "http://www.wowhead.com/?spells=7.11&filter=na=Teleport", },
                Process = ProcessDruidTeleport,
            },
            Home = {
                Spells = { "http://www.wowhead.com/?spells&filter=na=home%20location;ex=on", },
                Process = ProcessHomeTeleport,
            },
            Spells = { "http://www.wowhead.com/?spells=0&filter=na=Teleports+the+caster;ex=on", },
            Process = ProcessTeleport,
        },
    }, --Mounts
};
local convert = {};
local collections = { [latestVersion] = {} };
local latestCollections = collections[latestVersion];
local compatibleCollections = nil;
local prebuilt =  {
    ["PetAndMountDatabase.Items.MountEnchant"] = "21212,37750,37816,21213",
    ["PetAndMountDatabase.Items.CritterEnchant"] = "37460,44820,37460,35223,37431",
--    ["PetAndMountDatabase.Spells.Teleport.Items"] = "-54406:40585|40586,-8690:6948,-46149:35230,-39937:28585",
    ["PetAndMountDatabase.Spells.Teleport.Items.Engineering.Gnomish"] = "-23452:18986,-30544:30544",
    ["PetAndMountDatabase.Spells.Teleport.Items.Engineering.Goblin"] = "-36890:30542,-23442:18984",
    ["PetAndMountDatabase.MiniPets.Children"] = "m,PetAndMountDatabase.Critters.Children",
    ["PetAndMountDatabase.MiniPets.Equipment.Head"] = "m,PetAndMountDatabase.Critters.Equipment.Head",
    ["PetAndMountDatabase.MiniPets.Normal"] = "m,PetAndMountDatabase.Critters.Normal",
    ["PetAndMountDatabase.MiniPets.Quest"] = "m,PetAndMountDatabase.Critters.Quest",
    ["PetAndMountDatabase.MiniPets.Reagented.Snowball"] = "m,PetAndMountDatabase.Critters.Reagented.Snowball",
};

if( compatiblityVersion ) then
    data[compatiblityVersion] = persistence.load( ("datacache.%i.lua"):format( compatiblityVersion ) ) or { database = {}, };
    compatibleData = data[compatiblityVersion];
    collections[compatiblityVersion] = {};
    compatibleCollections = collections[compatiblityVersion];
end--if

function GetMountSpeed( page, mount )
    local flyingSpeed = page:match( "Apply Aura: Mod Speed Mounted Flight<small><br />Value: (%d+)</small>" );
    local groundSpeed = page:match( "Apply Aura: Mod Speed Mounted<small><br />Value: (%d+)</small>" );
    local aquaticSpeed = page:match( "Apply Aura: Mod Speed Swim<small><br />Value: (%d+)</small>" );
    local flying = flyingSpeed or page:find( "[Cc]an only be [%w]+ed in [%w%s]-." );
    local aquatic = aquaticSpeed;
    local passengers = page:match( "can carry (%d+) passengers" ) or ( ( page:match( "complete with two buddies[%.!]" ) or page:match( "Allows [%w%s]- passengers to board." ) ) and 3 ) or ( page:match( "%[Summon%]&nbsp;%(<a href=\"/%?npc=%d+\">%d+</a>%)<small><br />Value: %d+</small>" ) and 2 ) or ( page:match( "Rides and parks " ) and 2 );
    local variable = page:find( "This [%w%s]- changes [%w%s]-depending on your Riding skill[%w%s]-." );

    mount.Passengers = passengers and tonumber( passengers ) or 1;
    if( flyingSpeed ) then
        flyingSpeed = tonumber( flyingSpeed ) + 100;
    else
        flyingSpeed = 0;
    end--if
    if( aquaticSpeed ) then
        aquaticSpeed = tonumber( aquaticSpeed ) + 100;
    elseif( flyingSpeed > 0 ) then
        aquaticSpeed = 0;
    else
        aquaticSpeed = 100;
    end--if
    if( groundSpeed ) then
        groundSpeed = tonumber( groundSpeed ) + 100;
    else
        groundSpeed = 100;
    end--if
    if( flying ) then
        if( variable ) then
            mount.Type[1] = "Variable";
            mount.Type[2] = "Flying";
            flyingSpeed = -1;
            groundSpeed = -1;
            aquaticSpeed = 0;
        else
            mount.Type[1] = "Flying";
            if( flyingSpeed == 410 ) then
                mount.Type[2] = "ExtremelyFast";
            elseif( flyingSpeed == 380 ) then
                mount.Type[2] = "VeryFast";
            else
                mount.Type[2] = "Fast";
            end--if
        end--if
    elseif( aquatic ) then
        mount.Type[1] = "Aquatic";
        if( aquaticSpeed == 200 ) then
            mount.Type[2] = "VeryFast";
        elseif( aquaticSpeed == 160 ) then
            mount.Type[2] = "Fast";
        else
            mount.Type[2] = "Slow";
        end--if
    elseif( variable ) then
        mount.Type[1] = "Variable";
        mount.Type[2] = "All";
        flyingSpeed = -1;
        groundSpeed = -1;
        aquaticSpeed = 100;
    else
        mount.Type[1] = "Ground";
        if( groundSpeed == 200 ) then
            mount.Type[2] = "VeryFast";
        elseif( groundSpeed == 160 ) then
            mount.Type[2] = "Fast";
        else
            mount.Type[2] = "Slow";
        end--if    
    end--if
    mount.GroundSpeed = groundSpeed;
    mount.AquaticSpeed = aquaticSpeed;
    mount.FlyingSpeed = flyingSpeed;
end--GetMountSpeed( page, mount )

function ReagentedCheck( page, item )
    local reagent, name = page:match( "Reagents: .-<a href=\"/%?item=(%d+)\">(.-)</a>" );
    
    if( reagent ) then
        item.Type[1] = "Reagented";
        item.Type[2] = name;
        item.Value = reagent;
    end--if
end--ReagentedCheck( page, item )

function IgnoreSpell( page )
    return page:find( "Summons a .+ that will protect you for %d+ min." ) or page:find( "Summons a .+ to your aid for %d+ sec." );
end--IngnoreSpell( page )

function ProcessSpellList( url, dest, mountCheck, Process )
    local page = GetPage( url );
    local prefix = url:match( "http://(%w+)%." );

    if( page ) then
        for spellID, name in page:gmatch( "<a href=\"/%?spell=(%d+)\">(.-)</a>" ) do
            spellID = tonumber( spellID );
            if( not dest[spellID] ) then
                page = GetPage( ("http://%s.wowhead.com/?spell=%i"):format( prefix, spellID ) );
                if( page ) then
                    if( not IgnoreSpell( page ) ) then
                        print( ("Found spell %s (%i)!"):format( name, spellID ) );
                        dest[spellID] = { Spell = spellID, SpellName = name, Type = {}, };
                        if( mountCheck ) then
                            GetMountSpeed( page, dest[spellID] );
                        end--if
                        ReagentedCheck( page, dest[spellID] );
                        for teaches, data in page:gmatch( "%({template: 'item', id: '(%l+)%-by%-item', name: LANG%.tab_%l+, tabs: tabsRelated, parent: 'lkljbjkb574', data: %[(.+)%]}%);" ) do
                            data = data .. ",";
                            for itemID, itemName, slot in data:gmatch( "{id:(%d+),name:'%d(.-)',.-,slot:(%d+),.-}," ) do
                                slot = tonumber( slot );
                                itemID = tonumber( itemID );
                                itemName = itemName:gsub( "\\", "" );
                                print( ("Found item %s [%s] for %s [%i]!"):format( itemName, itemID, name, spellID ) );
                                dest[spellID].ItemName = itemName;
                                dest[spellID].Item = itemID;
                                if( slot ~= 0 ) then
                                    if( slot == 1 ) then
                                        dest[spellID].InventorySlot = "HEADSLOT";
                                    end--if
                                end--if
                                if( teaches == "taught" ) then
                                    dest[spellID].Taught = true;
                                end--if
                                page = GetPage( ("http://%s.wowhead.com/?item=%i"):format( prefix, itemID ) );
                                if( page ) then
                                    local riding = page:match( "Requires <a href=\"/%?spells=9%.762\" class=\"q1\">Riding</a> %((%d+)%)" );
                                    local engineering = page:match( "Requires <a href=\"/%?spells=11%.202\" class=\"q1\">Engineering</a> %((%d+)%)" ) or page:match( "Requires (%d+) engineering skill to ride." );
                                    local tailoring = page:match( "Requires <a href=\"/%?spells=11%.197\" class=\"q1\">Tailoring</a> %((%d+)%)" ) or page:match( "Requires (%d+) skill in tailoring to ride." );

                                    if( not riding ) then
                                        riding = page:match( "Requires <a href=\"/%?spell=%d+\" class=\"q1\">(%w+) Riding</a>" );

                                        if( riding == "Artisan" ) then
                                            riding = 300;
                                        elseif( riding == "Expert" ) then
                                            riding = 225;
                                        elseif( riding == "Journeyman" ) then
                                            riding = 150;
                                        elseif( riding == "Apprentice" ) then
                                            riding = 75;
                                        else
                                            riding = nil;
                                        end--if
                                    end--if
                                    if( riding ) then
                                        if( not dest[spellID].Skills ) then
                                            dest[spellID].Skills = {};
                                        end--if
                                        dest[spellID].Skills.Riding = tonumber( riding );
                                    end--if
                                    if( engineering ) then
                                        if( not dest[spellID].Skills ) then
                                            dest[spellID].Skills = {};
                                        end--if
                                        dest[spellID].Skills.Engineering = tonumber( engineering );
                                    end--if
                                    if( tailoring ) then
                                        if( not dest[spellID].Skills ) then
                                            dest[spellID].Skills = {};
                                        end--if
                                        dest[spellID].Skills.Tailoring = tonumber( tailoring );
                                    end--if
                                end--if
                            end--for
                        end--if
                        if( dest[spellID].Item ) then
                            if( type( dest[spellID].Item ) == "table" ) then
                                for _, itemID in pairs( dest[spellID].Item ) do
                                    reverseLookup[itemID] = spellID;
                                end--for
                            else
                                reverseLookup[dest[spellID].Item] = spellID;
                            end--if
                        end--if
                        if( Process ) then
                            Process( dest[spellID] );
                        end--if
                    end--if
                end--if
            end--if
        end--for
    else
        print( ("Could not open url %s"):format( url ) );
    end--if
end--ProcessSpellList( url, dest, mountCheck, Process )

function ProcessItemsList( url, dest, mountCheck, Process )
    local page = GetPage( url );
    local prefix = url:match( "http://(%w+)%." );

    if( page ) then
        for itemID, name in page:gmatch( "<a href=\"/%?item=(%d+)\">(.-)</a>" ) do
            itemID = tonumber( itemID );
            if( not reverseLookup[itemID] ) then
                local itemName = name;

                page = GetPage( ("http://%s.wowhead.com/?item=%i"):format( prefix, itemID ) );
                if( page ) then
                    local riding = page:match( "Requires <a href=\"/%?spells=9%.762\" class=\"q1\">Riding</a> %((%d+)%)" );
                    local engineering = page:match( "Requires <a href=\"/%?spells=11%.202\" class=\"q1\">Engineering</a> %((%d+)%)" ) or page:match( "Requires (%d+) engineering skill to ride." );
                    local tailoring = page:match( "Requires <a href=\"/%?spells=11%.197\" class=\"q1\">Tailoring</a> %((%d+)%)" ) or page:match( "Requires (%d+) skill in tailoring to ride." );
                    local spellID = page:match ( "{template: 'spell', id: 'teaches%-ability',.- data: %[{id:(%d+)," );
                    local taught = nil;

                    if( not riding ) then
                        riding = page:match( "Requires <a href=\"/%?spell=%d+\" class=\"q1\">(%w+) Riding</a>" );

                        if( riding == "Artisan" ) then
                            riding = 300;
                        elseif( riding == "Expert" ) then
                            riding = 225;
                        elseif( riding == "Journeyman" ) then
                            riding = 150;
                        elseif( riding == "Apprentice" ) then
                            riding = 75;
                        else
                            riding = nil;
                        end--if
                    end--if
                    if( not spellID ) then
                        spellID = page:match ( "<a href=\"/%?spell=(%d+)\" class=\"q2\">" );
                    else
                        taught = true;
                    end--if
                    if( spellID ) then
                        spellID = tonumber( spellID );
                        page = GetPage( ("http://%s.wowhead.com/?spell=%i"):format( prefix, spellID ) );
                        if( page ) then
                            if( not IgnoreSpell( page ) ) then
                                name = page:match( "<title>(.+) %- Spell %- World of Warcraft</title>" ) or name;
                                if( not dest[spellID] ) then
                                    dest[spellID] = { Spell = spellID, SpellName = name, Taught = taught, Type = {}, };
                                    if( mountCheck ) then
                                        GetMountSpeed( page, dest[spellID] );
                                    end--if
                                    ReagentedCheck( page, dest[spellID] );
                                    for teaches, data in page:gmatch( "%({template: 'item', id: '(%l+)%-by%-item', name: LANG%.tab_%l+, tabs: tabsRelated, parent: 'lkljbjkb574', data: %[(.+)%]}%);" ) do
                                        data = data .. ",";
                                        for itemID, itemName, slot in data:gmatch( "{id:(%d+),name:'%d(.-)',.-,slot:(%d+),.-}," ) do
                                            itemID = tonumber( itemID );
                                            slot = tonumber( slot );
                                            itemName = itemName:gsub( "\\", "" );
                                            if( teaches == "taught" ) then
                                                dest[spellID].Taught = true;
                                            end--if
                                            if( slot ~= 0 ) then
                                                if( slot == 1 ) then
                                                    dest[spellID].InventorySlot = "HEADSLOT";
                                                end--if
                                            end--if
                                            print( ("Found item %s [%i] for %s [%i]!"):format( itemName, itemID, name, spellID ) );
                                            if( dest[spellID].Item ) then
                                                if( type( dest[spellID].Item ) == "table" ) then
                                                    dest[spellID].Item[#dest[spellID].Item + 1] = itemID;
                                                else
                                                    dest[spellID].Item = {
                                                        [1] = dest[spellID].Item,
                                                        [2] = itemID,
                                                    };
                                                end--if
                                            else
                                                dest[spellID].Item = itemID;
                                            end--if
                                        end--for
                                    end--if
                                    if( Process ) then
                                        Process( dest[spellID] );
                                    end--if
                                end--if
                                dest[spellID].ItemName = itemName;
                                dest[spellID].Item = itemID;
                                dest[spellID].Taught = taught
                                if( riding ) then
                                    if( not dest[spellID].Skills ) then
                                        dest[spellID].Skills = {};
                                    end--if
                                    dest[spellID].Skills.Riding = tonumber( riding );
                                end--if
                                if( engineering ) then
                                    if( not dest[spellID].Skills ) then
                                        dest[spellID].Skills = {};
                                    end--if
                                    dest[spellID].Skills.Engineering = tonumber( engineering );
                                end--if
                                if( tailoring ) then
                                    if( not dest[spellID].Skills ) then
                                        dest[spellID].Skills = {};
                                    end--if
                                    dest[spellID].Skills.Tailoring = tonumber( tailoring );
                                end--if
                                reverseLookup[itemID] = spellID;                                
                            end--if
                        end--if
                    end--if
                end--if
            end--if
        end--for
    else
        print( ("Could not open url %s"):format( url ) );
    end--if
end--ProcessSpellList( url, dest )

function ProcessList( major, list, nonmajor )
    for _, url in pairs( list.Spells ) do
        print( ("\nLooking for %s spells..."):format( major ) );
        ProcessSpellList( url, latestData.database[major], major == "Mounts", nonmajor );
    end--for
    if( list.Items ) then
        for _, url in pairs( list.Items ) do
            print( ("\nLooking for %s items..."):format( major ) );
            ProcessItemsList( url, latestData.database[major], major == "Mounts", nonmajor );
        end--for
    end--if
end--ProcessList( major, list )

function ProcessSub( major, sub, nonmajor )
    if( sub.Spells ) then
        ProcessList( major, sub, nonmajor and sub.Process or nil );
    end--if
    for name, info in pairs( sub ) do
        if( ( name ~= "Spells" ) and ( name ~= "Items" ) and ( name ~= "ProcessList" ) and ( type( info ) == "table" ) ) then
            ProcessSub( major, info, true );
        end--if
    end--if
end--ProcessSub( sub )

for major, sets in pairs( latestData.database ) do
    for _, id in pairs( sets ) do
        if( id.Item ) then
            if( type( id.Item ) == "table" ) then
                for _, itemID in pairs( id.Item ) do
                    reverseLookup[itemID] = id.Spell;
                end--for
            else
                reverseLookup[id.Item] = id.Spell;
            end--if
        end--if
    end--for
end--for
for major, subs in pairs( urlLists ) do
    if( not latestData.database[major] ) then
        latestData.database[major] = {};
    end--if
    ProcessSub( major, subs );
end--for

function AddToCollections( collectionsString, id, idUse )
    if( collectionsString ) then
        if( id.Value ) then
            return ("%s,%i:%s"):format( collectionsString, idUse, id.Value );
        else
            return ("%s,%i"):format( collectionsString, idUse );
        end--if
    else
        if( id.Value ) then
            return ("%i:%s"):format( idUse, id.Value );
        else
            return ("%i"):format( idUse );
        end--if
    end--if
end--AddToCollections( collectionsString, id, idUse )

for version, database in pairs( data ) do
    for major, sets in pairs( database.database ) do
        for _, id in pairs( sets ) do
            if( urlLists[major] and urlLists[major].Process ) then
                urlLists[major].Process( id );
            end--if
    
            if( type( id.Type ) == "string" ) then
                id.Type = { [1] = id.Type, [2] = id.Subtype, [3] = id.Subtype2, };
                id.Subtype = nil;
                id.Subtype2 = nil;
            end--if

            if( ( id.Item or id.ClassSpell ) and not( id.Ignore ) ) then
                local collection = major;
                local idUse = ( id.Taught or id.ClassSpell or ( not id.Item ) ) and id.Spell * -1 or id.Item;
    
                for i, typestr in pairs( id.Type ) do
                    collection = collection .. "." .. typestr;
                end--for
                if( not convert[major] ) then
                    convert[major] = {};
                end--if
                convert[major][id.Spell] = id;

                if( type( idUse ) == "table" ) then
                    for _, idUse in pairs( idUse ) do
                        collections[version][collection] = AddToCollections( collections[version][collection], id, idUse );
                    end--if
                else
                    collections[version][collection] = AddToCollections( collections[version][collection], id, idUse );
                end--if
                if( id.Family ) then
                    for _, class in pairs( id.Family ) do
                        collection = ("Family.%s.%s"):format( major, class );
						if( type( idUse ) == "table" ) then
							for _, idUse in pairs( idUse ) do
								collections[version][collection] = AddToCollections( collections[version][collection], id, idUse );
							end--if
						else
							collections[version][collection] = AddToCollections( collections[version][collection], id, idUse );
						end--if						
                    end--for
                end--if
                if( id.Source ) then
                    for _, class in pairs( id.Source ) do
                        collection = ("Source.%s.%s"):format( major, class );
						if( type( idUse ) == "table" ) then
							for _, idUse in pairs( idUse ) do
								collections[version][collection] = AddToCollections( collections[version][collection], id, idUse );
							end--if
						else
							collections[version][collection] = AddToCollections( collections[version][collection], id, idUse );
						end--if						
                    end--for
                end--if
            end--if
        end--for
    end--for
end--for

print( "\nWriting database to file..." );

persistence.store( "datacache.lua", data[latestVersion] );
if( compatiblityVersion ) then
    persistence.store( ("datacache.%i.lua"):format( compatiblityVersion ), data[compatiblityVersion] );
end--if

local hFile = io.open( "LibLordFarlander-PeriodicTable-3.1-PetAndMountDatabase-2.0.lua", "w+" );

if( hFile == nil ) then
    return;
end--if

function WriteSet( collection )
    hFile:write( "    LibStub( \"LibPeriodicTable-3.1\" ):AddData( \"PetAndMountDatabase\", tostring( tonumber( (\"$Rev: 164 $\"):match( \"%d+\" ) ) + 90000 ), {\n" );
    for setName, set in pairs( collection ) do
        hFile:write( ("        [\"PetAndMountDatabase.%s\"] = %q,\n"):format( setName, set ) );
    end--for
    for setName, set in pairs( prebuilt ) do
        hFile:write( ("        [%q] = %q,\n"):format( setName, set ) );
    end--for

    local function AddItem( writestring, id, item )
        if( writestring ) then
            return ("%s,%i:%i"):format( writestring, item, id.Spell * -1 );
        else
            return ("%i:%i"):format( item, id.Spell * -1 );
        end--if
    end--AddItem( writestring, id )

    for major, sets in pairs( convert ) do
        local writestring;

        for _, id in pairs( sets ) do
            if( id.Spell and ( not id.Ignore ) and ( not id.ClassSpell ) ) then
                if( type( id.Item ) == "table" ) then
                    for _, item in pairs( id.Item ) do
                        writestring = AddItem( writestring, id, item );
                    end--for
                else
                    writestring = AddItem( writestring, id, id.Item );
                end--if
            end--if
        end--for
        if( writestring ) then
            hFile:write( ("        [\"PetAndMountDatabase.Conversion.%s\"] = %q,\n"):format( major, writestring ) );
        end--if
    end--for
    for setName, set in pairs( prebuilt ) do
        hFile:write( ("        [%q] = %q,\n"):format( setName, set ) );
    end--for
    hFile:write( "    } );\n" );
end--

hFile:write( HEADER );
if( compatiblityVersion ) then
    hFile:write( ("if( tonumber( select( 4, GetBuildInfo() ) ) >= %i ) then\n"):format( compatiblityVersion ) );
    WriteSet( compatibleCollections );
    hFile:write( "else\n" );
end--if
WriteSet( latestCollections );
if( compatiblityVersion ) then
    hFile:write( "end--if\n" );
end--if
hFile:close();

--tempdebug
-- local expanded = {}
-- for major, sets in pairs( latestData.database ) do
--     if( not expanded[major] ) then
--         expanded[major] = {};
--     end--if
--     for name, id in pairs( sets ) do
--         if( id.Spell and ( id.Item or id.ClassSpell ) and ( not id.Ignore ) ) then
--             local putinto = nil;
--             local ids = id.Spell * -1;
--     
--             if( not id.Taught and id.Item ) then
--                 ids = id.Item;
--             end--if
--             if( not expanded[major][id.Type] ) then
--                 expanded[major][id.Type] = {};
--             end--if
--             putinto = expanded[major][id.Type];
--             if( id.Subtype ) then
--                 if( not expanded[major][id.Type][id.Subtype] ) then
--                     expanded[major][id.Type][id.Subtype] = {};
--                 end--if
--                 putinto = expanded[major][id.Type][id.Subtype];
--             end--if
--             if( id.Subtype1 ) then
--                 if( not expanded[major][id.Type][id.Subtype][id.Subtype1] ) then
--                     expanded[major][id.Type][id.Subtype][id.Subtype1] = {};
--                 end--if
--                 putinto = expanded[major][id.Type][id.Subtype][id.Subtype1];
--             end--if
--             putinto[ids] = { value = id.Value, name = id.SpellName };
--         end--if
--     end--for
-- end--for
-- 
-- persistence.store( "debug.lua", expanded );